(***********************************************************************)
(*                                                                     *)
(*                          Sklml                                      *)
(*                                                                     *)
(*             Pierre Weis (Pierre.Weis@inria.fr)                      *)
(*             Francois Clement (Francois.Clement@inria.fr)            *)
(*             Quentin Carbonneaux (Quentin.Carbonneaux@inria.fr)      *)
(*                                                                     *)
(* This program is free software.                                      *)
(* (see the file LICENSE in the root directory of the distribution).   *)
(*                                                                     *)
(***********************************************************************)

(* $Id: sklstream.ml,v 1.6 2010-12-02 10:12:58 weis Exp $ *)

module type SOW = sig
(** This is the core of the stream feature for Sklml. *)
  type 'a wire_data =(* private *)
    | Data of 'a
    | Begin_sub_stream
    | End_sub_stream
  ;;
  type 'a cell = private
    | Nil
    | Cons of 'a wire_data * 'a stream

  and 'a stream = 'a cell Lazy.t
  ;;

  val nil_cell : 'a cell;;
  val cons_wire_data_cell : 'a wire_data * 'a stream -> 'a cell;;
  val cons_data_cell : 'a * 'a stream -> 'a cell;;
  val cons_begin_cell : 'a stream -> 'a cell;;
  val cons_end_cell : 'a stream -> 'a cell;;
  val get_cell : 'a stream -> 'a cell;;
  val nil : 'a stream;;
  val cons : 'a * 'a stream -> 'a stream;;
  val cons_sub_stream : 'a stream * 'a stream -> 'a cell;;
  val append : 'a stream -> 'a stream -> 'a stream;;
end
;;

module Stream_on_wire : SOW = struct
  type 'a wire_data =
    | Data of 'a
    | Begin_sub_stream
    | End_sub_stream
  ;;

  type 'a cell =
    | Nil
    | Cons of 'a wire_data * 'a stream

  and 'a stream = 'a cell Lazy.t
  ;;

  let nil_cell = Nil
  and cons_begin_cell s = Cons (Begin_sub_stream, s)
  and cons_end_cell s = Cons (End_sub_stream, s)
  and cons_data_cell (x, s) = Cons (Data x, s)
  and cons_wire_data_cell (d, s) = Cons (d, s)
  ;;

  let get_cell = Lazy.force;;

  (* Stream construction functions. *)
  let nil = lazy Nil;;
  let cons (x, s) = lazy (cons_data_cell (x, s));;

  let cons_sub_stream (sub, s) =

    let rec loop sub =
      match get_cell sub with
      | Nil -> cons_end_cell s
      | Cons (d, sub) -> cons_wire_data_cell (d, lazy (loop sub)) in

    cons_begin_cell (lazy (loop sub))
  ;;

  let rec append s1 s2 =

    let in_append s1 =
      match get_cell s1 with
      | Nil -> get_cell s2
      | Cons (d, s1) -> cons_wire_data_cell (d, append s1 s2) in

    lazy (in_append s1)
  ;;
end
;;

open Stream_on_wire;;

exception End_of_stream;;

type 'a stream = 'a Stream_on_wire.stream;;

type 'a t = 'a stream;;

(* Generate a lazy stream from a function. *)
let rec of_fun =

  let in_fun f =
    let h =
      try Some (f ()) with
      | End_of_stream -> None in
    match h with
    | Some x -> cons (x, of_fun f)
    | None -> nil in

  fun f -> lazy (get_cell (in_fun f))
;;

(* Generate a lazy stream from a function. *)
let rec of_fun =

  let in_fun f =
    match f () with
    | Some x -> cons (x, of_fun f)
    | None -> nil in

  fun f -> lazy (get_cell (in_fun f))
;;

(* Generate a lazy stream from a list. *)
let rec of_list =

  let in_of_list = function
    | [] -> nil
    | x :: l -> cons (x, of_list l) in

  fun l -> lazy (get_cell (in_of_list l))
;;

(* Generate a stream from a vector. *)
let of_vect v = of_list (Array.to_list v);;

(* Generate the list of elements of a lazy stream.
   Warning: this function is eager, meaning that the stream argument is
   completely evaluated (if this is possible, otherwise the function loops for
   ever). *)
let to_list s =
  let rec loop accu t =
    match get_cell t with
    | Nil -> List.rev accu
    | Cons (End_sub_stream, t)
    | Cons (Begin_sub_stream, t) -> loop accu t
    | Cons (Data x, t) -> loop (x :: accu) t in
  loop [] s
;;

(* Generate the vector of elements of a lazy stream.
   Warning: this function is eager, meaning that the stream argument is
   completely evaluated (if this is possible, otherwise the function loops
   for ever). *)
let to_vect s = Array.of_list (to_list s);;

(* Vector and list serialization. *)
let serialize_list (l, s) = cons_sub_stream (of_list l, s);;
let serialize_vector (v, s) = serialize_list (Array.to_list v, s);;

let unserialize_list s =

  let rec get_list accu s =
    match get_cell s with
    | Nil
    | Cons (Begin_sub_stream, _) -> assert false
    | Cons (Data x, t) -> get_list (x :: accu) t
    | Cons (End_sub_stream, t) -> List.rev accu, t in

  get_list [] s
;;

let unserialize_vector s =
  let l, s = unserialize_list s in
  Array.of_list l, s
;;

let rec serialize_stream_vector =

  let rec in_serialize_stream_vector s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Data v, t) ->
      serialize_vector (v, serialize_stream_vector t)
    | Cons (Begin_sub_stream, t) ->
      cons_begin_cell (lazy (end_critical t))
    | Cons (End_sub_stream, t) ->
      cons_end_cell (serialize_stream_vector t)

  and end_critical s =
    match get_cell s with
    | Nil -> assert false
    | Cons ((Begin_sub_stream | Data _), _) ->
      in_serialize_stream_vector s
    | Cons (End_sub_stream, _) ->
      serialize_list ([], serialize_stream_vector s) in

  fun s -> lazy (in_serialize_stream_vector s)
;;

let of_sarray = serialize_stream_vector;;

let unserialize_stream_vector =

  let rec in_unserialize_stream_vector s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Begin_sub_stream, t) ->
      after_begin t
    | Cons ((End_sub_stream | Data _), _) ->
      assert false

  and after_begin s =
    match get_cell s with
    | Nil -> assert false
    | Cons ((End_sub_stream | Data _), _) ->
      let v, t = unserialize_vector s in
      cons_data_cell (v, lazy (after_end t))
    | Cons (Begin_sub_stream, t) ->
      cons_begin_cell (lazy (after_begin t))

  and after_end s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Begin_sub_stream, t) ->
      after_begin t
    | Cons (End_sub_stream, t) ->
      cons_end_cell (lazy (after_end t))
    | Cons (Data _, _) -> assert false in

  fun s -> lazy (in_unserialize_stream_vector s)
;;
let to_sarray = unserialize_stream_vector;;

(* [map f s] maps the function [f] onto the stream [s]:
   if [s] is [x1; x2; ...; xn], then [map f s] returns the stream
   [f x1; f x2; ...; f xn].
*)
let rec map f =

  let in_map s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Begin_sub_stream, t) ->
      cons_begin_cell (map f t)
    | Cons (End_sub_stream, t) ->
      cons_end_cell (map f t)
    | Cons (Data x, t) ->
      cons_data_cell (f x, map f t) in

  fun s -> lazy (in_map s)
;;
let lift = map;;

let foldl f z =

  let rec in_foldl s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Begin_sub_stream, t) ->
      after_begin t
    | Cons (End_sub_stream, _) ->
      assert false
    | Cons (Data _, _) ->
      assert false

  and after_begin s =
    match get_cell s with
    | Nil -> assert false
    | Cons (Begin_sub_stream, t) ->
      cons_begin_cell (lazy (after_begin t))
    | Cons ((End_sub_stream | Data _), _) ->
      accu_foldl z s

  and accu_foldl z s =
    match get_cell s with
    | Nil -> assert false
    | Cons (Begin_sub_stream, _) ->
      assert false
    | Cons (End_sub_stream, t) ->
      cons_data_cell (z, lazy (after_end t))
    | Cons (Data x, t) ->
      accu_foldl (f z x) t

  and after_end s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Begin_sub_stream, t) ->
      after_begin t
    | Cons (End_sub_stream, t) ->
      cons_end_cell (lazy (after_end t))
    | Cons (Data _, _) -> assert false in

  fun s -> lazy (in_foldl s)
;;

let singleton x = cons (x, nil);;

let head s =
  match get_cell s with
  | Nil ->
    failwith "head: premature end of stream"
  | Cons (Begin_sub_stream, _) ->
    failwith "head: unexpected beginning of sub stream"
  | Cons (End_sub_stream, _) ->
    failwith "head: unexpected end of sub stream"
  | Cons (Data x, _) -> x
;;

(* Given a stream starting with a (presumably single) sub stream having a
  (presumably single element), [head_sub_stream] returns this element.
  See [unlift] below for details. *)
let head_sub_stream =

  let rec head_sub_stream s =
    match get_cell s with
    | Nil ->
      failwith "head: premature end of stream while looking for beginning of sub stream"
    | Cons (Begin_sub_stream, t) -> head_sub_stream_elem t
    | Cons (End_sub_stream, _) ->
      failwith "head: premature end of sub stream while looking for beginning of sub stream"
    | Cons (Data _, _) ->
      failwith "head: premature element while looking for beginning of sub stream"

  and head_sub_stream_elem t =
    match get_cell t with
    | Nil ->
      failwith "head: premature end of stream  while looking for element"
    | Cons (Begin_sub_stream, _) ->
      failwith "head: found begin of sub stream while looking for element"
    | Cons (End_sub_stream, _) ->
      failwith "head: found end of sub stream while looking for element"
    | Cons (Data x, t) -> head_sub_stream_end x t

  and head_sub_stream_end x t =
    match get_cell t with
    | Nil ->
      failwith "head: premature end of stream while looking for end of sub stream"
    | Cons (Begin_sub_stream, _) ->
      failwith "head: found begin of sub stream while looking for end of sub stream"
    | Cons (End_sub_stream, _) -> x
    | Cons (Data _, _) ->
      failwith "head: Found a data while looking for end of sub stream" in

  fun s -> head_sub_stream s
;;

let unlift skl =
  fun x ->
  head_sub_stream (skl (lazy (cons_sub_stream (singleton x, nil))))
;;

let fork = fun s ->

  let rec fork_cell s =
    match get_cell s with
    | Nil -> nil, nil
    | Cons (x, t) ->
      let pair = lazy (fork_cell t) in
      lazy (cons_wire_data_cell (x, lazy (get_cell (snd (Lazy.force pair))))),
      lazy (get_cell (fst (Lazy.force pair))) in

  let rec double_control s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons ((Begin_sub_stream | End_sub_stream) as h, t) ->
      cons_wire_data_cell (h, lazy (
      cons_wire_data_cell (h, lazy (
      double_control t))))
    | Cons (Data _ as h, t) ->
      cons_wire_data_cell (h, lazy (double_control t)) in

  let pair = lazy (fork_cell (lazy (double_control s))) in
  lazy (get_cell (fst (Lazy.force pair))),
  lazy (get_cell (snd (Lazy.force pair)))
;;

(* Merges two lazy stream into a single lazy stream with intermixed elements
   from both streams in a fair way (alternatively from one stream at a time).
   Note that the two streams must not have the same number of elements: in
   case one stream is exhausted, the result is the other (the longer one). *)
let join =

  let rec in_join_cell (s1, s2) =
    match get_cell s1 with
    | Nil -> get_cell s2
    | Cons (x, t) -> cons_wire_data_cell (x, lazy (in_join_cell (s2, t))) in

  let rec single_control s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons ((Begin_sub_stream | End_sub_stream), t) ->
      send_control t
    | Cons (Data _ as h, t) ->
      cons_wire_data_cell (h, lazy (single_control t))
  and send_control s =
    match get_cell s with
    | Cons ((Begin_sub_stream | End_sub_stream) as h, t) ->
      cons_wire_data_cell (h, lazy (single_control t))
    | Cons (Data _, _) -> assert false
    | Nil -> assert false in

  fun (s1, s2) -> lazy (single_control (lazy (in_join_cell (s1, s2))))
;;

(* [iter f s] iterates the function [f] on the stream [s]:
   if [s] is [x1; x2; ...; xn], then [iter f s] performs the sequence
   [f x1; f x2; ...; f xn].
   Note: any call to [iter] completely unfreeezes its the stream argument. *)
let rec iter f s =
  match get_cell s with
  | Nil -> ()
  | Cons (Data x, t) -> f x; iter f t
  | Cons (End_sub_stream, t)
  | Cons (Begin_sub_stream, t) -> iter f t
;;

(* LOOP FUNCTIONS *)

let rec filter p =

  let in_filter s1 s2 =
    match get_cell s1, get_cell s2 with
    | Cons (Data x, t1), Cons (Data y, t2) ->
      if p x
      then cons_data_cell (y, filter p t1 t2)
      else get_cell (filter p t1 t2)
    | Cons (Begin_sub_stream, t1), Cons (Begin_sub_stream, t2) ->
      cons_begin_cell (filter p t1 t2)
    | Cons (End_sub_stream, t1), Cons (End_sub_stream, t2) ->
      cons_end_cell (filter p t1 t2)
    | Nil, Nil -> nil_cell
    | (Nil | Cons ((Begin_sub_stream | End_sub_stream | Data _), _)),
      (Nil | Cons ((Begin_sub_stream | End_sub_stream | Data _), _)) ->
      assert false in

  fun sb s -> lazy (in_filter sb s)
;;

let with_ht s def func =
  match get_cell s with
  | Cons (hd, tl) -> func hd tl
  | Nil -> def
;;

let rec get_fst_data s =
  with_ht s (fun _ -> nil_cell) (fun hd tl ->
    match hd with
    | Begin_sub_stream ->
      fun k -> cons_begin_cell (lazy ((get_fst_data tl) k))
    | End_sub_stream ->
      fun k -> cons_end_cell (lazy ((get_fst_data tl) k))
    | Data x ->
      fun k -> cons_data_cell (x, lazy (k tl))
  )
;;

let loop_core lb ls1 s2 =

  let rec in_loop_core sb s1 s2 =
    match get_cell sb, get_cell s1 with
    | Cons (Data b, tb), Cons (Data x1, t1) ->
      if b
      then cons_data_cell (x1, lazy (in_loop_core tb t1 s2))
      else get_fst_data s2 (in_loop_core tb t1)
    | Cons (Begin_sub_stream, tb), Cons (Begin_sub_stream, t1)
    | Cons (End_sub_stream, tb), Cons (End_sub_stream, t1) ->
      in_loop_core tb t1 s2
    | (Nil | Cons ((Begin_sub_stream | End_sub_stream | Data _), _)),
      (Nil | Cons ((Begin_sub_stream | End_sub_stream | Data _), _)) ->
      assert false in

  lazy (
    get_fst_data s2 (in_loop_core (Lazy.force lb) (Lazy.force ls1))
  )
;;

let rec loop p f is =
  let rec los1 = lazy (f (Lazy.force lis1))
  and los2 = lazy (p (Lazy.force lis2))
  and lis1 = lazy (loop_core los2 los1 is)
  and lis2 = lazy (Lazy.force los1) in

  lazy (get_cell (filter not (Lazy.force los2) (Lazy.force los1)))
;;

(* Rail versions of fork and join *)

(****************************************
(* Initial version with double and single control (although not fully
   implemented  *)
let fork_rail n s =

  let rec fork_cell i s =
    match get_cell s with
    | Nil -> Array.make n nil
    | Cons (x, t) ->

      let arr = lazy (fork_cell ((i + 1) mod n) t) in

      Array.init n (fun j ->
        if i = j then lazy (
          cons_wire_data_cell (x, lazy (get_cell ((Lazy.force arr).(i)))))
        else lazy (
          get_cell ((Lazy.force arr).(j))
       )
      ) in

  let double_control s = Lazy.force s in

  let arr = lazy (fork_cell 0 (lazy (double_control s))) in

  Array.init n (fun i -> lazy (
    get_cell ((Lazy.force arr).(i)))
  )
;;

let is_nil s =
  match get_cell s with
  | Nil -> true
  | Cons (_, _) -> false
;;

let array__for_all p v =
  let lim = Array.length v - 1 in
  let rec loop i =
    if i <= lim then p v.(i) && loop (i + 1) else true in
  loop 0
;;

let fork_rail n s =

  let sv = Array.make n nil in

  let rec fork_cell i s =

    match get_cell s with
    | Nil ->
      (* Assign nil to all streams of sv or simply verify that
         all the streams are now indeed empty *)
      if not (array__for_all is_nil sv) then assert false

    | Cons (Begin_sub_stream | End_sub_stream as bg, t) ->
      Array.iteri
        (fun j _ -> sv.(j) <- lazy (
          cons_wire_data_cell
            (bg, lazy (
             if j = i then fork_cell i t;
             get_cell (sv.(j))))))
        sv

    | Cons (Data _ as x, t) ->

      sv.(i) <- lazy (
        cons_wire_data_cell
          (x, lazy (
            fork_cell ((i + 1) mod n) t;
            get_cell (sv.(i)))))
  in

  sv.(0) <- lazy (
    fork_cell 0 s;
    get_cell (sv.(0)));

  sv
;;

(* Second version without single and double control:
   Begin and End sub stream are distributed to all the streams *)
let fork_rail n s =

  let rec fork_cell i s =
    match get_cell s with
    | Nil -> Array.make n nil

    | Cons (Begin_sub_stream | End_sub_stream as bg, t) ->
      let arr = lazy (fork_cell i t) in
      Array.init n (fun j -> lazy (
        cons_wire_data_cell (bg, lazy (
          get_cell ((Lazy.force arr).(j))
         )
        )
       )
      )

    | Cons (Data _ as x, t) ->
      let arr = lazy (fork_cell ((i + 1) mod n) t) in
      Array.init n (fun j ->
        if i = j then lazy (
          cons_wire_data_cell
            (x, lazy (get_cell ((Lazy.force arr).(j)))))
        else lazy (
          get_cell ((Lazy.force arr).(j))
       )
      ) in

  let arr = lazy (fork_cell 0 s) in

  Array.init n (fun i ->
    lazy (get_cell ((Lazy.force arr).(i))))
;;

(* Strict equality required! *)
let rec serialize_stream_vector_for_rail n =

  let rec in_serialize_stream_vector s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Data v, t) ->
      let len = Array.length v in
      if len <> n then
        failwith
          (Printf.sprintf
             "serialize_stream_vector_for_rail: \
              wrong vector of length %d, should have length %d" len n)
      else
      serialize_vector (v, serialize_stream_vector_for_rail n t)
    | Cons (Begin_sub_stream, t) ->
      cons_begin_cell (lazy (end_critical t))
    | Cons (End_sub_stream, t) ->
      cons_end_cell (serialize_stream_vector_for_rail n t)

  and end_critical s =
    match get_cell s with
    | Nil -> assert false
    | Cons ((Begin_sub_stream | Data _), _) ->
      in_serialize_stream_vector s
    | Cons (End_sub_stream, _) ->
      serialize_list ([], serialize_stream_vector_for_rail n s) in

  fun s -> lazy (in_serialize_stream_vector s)
;;

*************************************************)

(* Third version generalized to sparse matrices *)
let fork_rail n s =

  let rec fork_cell i s =
    match get_cell s with
    | Nil -> Array.make n nil

    | Cons (Begin_sub_stream | End_sub_stream as bg, t) ->
      let arr = lazy (fork_cell 0 t) in
      Array.init n (fun j -> lazy (
        cons_wire_data_cell (bg, lazy (
          get_cell ((Lazy.force arr).(j))
         )
        )
       )
      )

    | Cons (Data _ as x, t) ->
      let arr = lazy (fork_cell ((i + 1) mod n) t) in
      Array.init n (fun j ->
        if i = j then lazy (
          cons_wire_data_cell
            (x, lazy (get_cell ((Lazy.force arr).(j)))))
        else lazy (
          get_cell ((Lazy.force arr).(j))
       )
      ) in

  let arr = lazy (fork_cell 0 s) in

  Array.init n (fun i ->
    lazy (get_cell ((Lazy.force arr).(i))))
;;

let rec serialize_stream_vector_for_rail n =

  let rec in_serialize_stream_vector s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Data v, t) ->
      let len = Array.length v in
      if len > n then
        failwith
          (Printf.sprintf
             "serialize_stream_vector_for_rail: \
              wrong vector of length %d, should have length %d" len n)
      else
      serialize_vector (v, serialize_stream_vector_for_rail n t)
    | Cons (Begin_sub_stream, t) ->
      cons_begin_cell (lazy (end_critical t))
    | Cons (End_sub_stream, t) ->
      cons_end_cell (serialize_stream_vector_for_rail n t)

  and end_critical s =
    match get_cell s with
    | Nil -> assert false
    | Cons ((Begin_sub_stream | Data _), _) ->
      in_serialize_stream_vector s
    | Cons (End_sub_stream, _) ->
      serialize_list ([], serialize_stream_vector_for_rail n s) in

  fun s -> lazy (in_serialize_stream_vector s)
;;

let ignore_tag bg s =
  match get_cell s with
  | Nil -> assert false
  | Cons (x, t) ->
    if x = bg then t else assert false
;;

(* Verify that every stream in sv is nil *)
let check_nil sv =
  Array.iter (fun s ->
    match get_cell s with
    | Nil -> ()
    | Cons (_, _) ->
      assert false
  ) sv
;;

let join_rail sv =

  let n = Array.length sv in

  (* Preserves the argument *)
  let sv = Array.copy sv in

  let rec in_join_cell i =
    match get_cell sv.(i) with
    | Nil ->
      check_nil sv;
      nil_cell
    | Cons (Begin_sub_stream | End_sub_stream as bg, _) ->
      Array.iteri (fun i s -> sv.(i) <- ignore_tag bg s) sv;
      cons_wire_data_cell (bg, lazy (in_join_cell i))
    | Cons (Data _ as x, t) ->
      sv.(i) <- t;
      cons_wire_data_cell (x, lazy (in_join_cell ((i + 1) mod n))) in

  lazy (in_join_cell 0)
;;

let on_rails fv s =
  let n = Array.length fv in
  let sv = fork_rail n s in
  let fsv = Array.init n (fun i -> fv.(i) sv.(i)) in
  join_rail fsv
;;

let on_par_rails (f1, f2) =
  let fv = [| f1; f2 |] in
  on_rails fv
;;

let rail fv =
  let n = Array.length fv in
  fun s ->
   let vs = serialize_stream_vector_for_rail n s in
   let sv = fork_rail n vs in
   let fsv = Array.init n (fun i -> fv.(i) sv.(i)) in
   unserialize_stream_vector (join_rail fsv)
;;

(** Tiny test *)
let ssv = serialize_stream_vector;;

let ii s = iter ignore s;;

let rec nth n s =
  match get_cell s with
  | Nil -> assert false
  | Cons (x, t) ->
    if n = 0 then x else nth (n - 1) t
;;

let vs0 = of_list [ [| |]; [| 1; 2; 3; |]; [| 1; |]; ];;
let s0 = ssv vs0;;
let sv = fork_rail 3 s0;;
let s = join_rail sv;;
let vs = unserialize_stream_vector s;;
ii vs = ii vs0;;

let rail0 s = on_par_rails (lift succ, lift pred) s;;

let s1 = rail0 s0;;

ii (rail0 s1);;

let vs3 = of_list [
  [| 1; 2; 3; |];
  [| 4; 5; 6; |];
  [| 7; 8; 9; |];
]
;;

let rails s =
  let square i = i * i in
  on_rails [| lift succ; lift pred; lift square; |] s
;;

let s3 = ssv vs3;;
let rs3 = rails s3;;
let res3 = unserialize_stream_vector rs3;;

let wrong =
  let square i = i * i in
  rail [| lift succ; lift pred; lift square; |]
    (of_list [ [| |] ])
;;

(*
nth 0 wrong;;
Exception:
Failure
 "serialize_stream_vector_for_rail: wrong vector length 0, should be 3".
*)
